use std::path::Path;
use std::fs::File;
use std::io::{BufWriter, BufReader};

use serde::{Serialize, Deserialize};

use bincode::{serialize, deserialize, serialize_into, deserialize_from};

use sfml::graphics::{FloatRect, IntRect};
use sfml::system::Vector2;

use crate::{Error, Result};
use crate::engine::math::{clamp, Matrix};
use crate::engine::renderer::Drawable;
use crate::engine::texture::{Texture, byte_to_index, SPRITE_SIZE_F32, SPRITE_SIZE_I32};
use crate::engine::character::Damage;

/// This enum represents if the collision happens on the X axis or the Y axis.
#[derive(Copy, Clone)]
pub enum CollisionAxis {
    /// The X axis.
    X,

    /// The Y axis.
    Y,

    /// Both axis simultaneously
    Both,
}

impl CollisionAxis {
    /// Returns true if the collision occured on X axis.
    pub fn is_x(self) -> bool {
        match self {
            CollisionAxis::Y => false,
            _ => true,
        }
    }

    /// Returns true if the collision occured on Y axis.
    pub fn is_y(self) -> bool {
        match self {
            CollisionAxis::X => false,
            _ => true,
        }
    }
}

/// This struct represents the different sides from which a collision can occur.
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub struct CollisionTile {
    /// If the character comes from the top, it will collide if this bool is true.
    pub from_top: bool,

    /// If the character comes from the left, it will collide if this bool is true.
    pub from_left: bool,

    /// If the character comes from the right, it will collide if this bool is true.
    pub from_right: bool,

    /// If the character comes from the bottom, it will collide if this bool is true.
    pub from_bottom: bool,
}

impl CollisionTile {
    /// Creates a collision tile that does not collide.
    pub fn empty() -> CollisionTile {
        CollisionTile {
            from_top: false,
            from_left: false,
            from_right: false,
            from_bottom: false,
        }
    }

    /// Creates a collision tile that collides from every side.
    pub fn full() -> CollisionTile {
        CollisionTile {
            from_top: true,
            from_left: true,
            from_right: true,
            from_bottom: true,
        }
    }
}

/// This struct represents a renderable tile linking to its part in the tileset texture.
#[derive(Debug, Copy, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub struct GraphicTile(Option<i32>);

impl GraphicTile {


    /// Creates the correct graphic tile depending on the neighbours.
    ///
    /// A none will be considered solid.
    pub fn from_neighbour_options(
        top_left: Option<CollisionTile>,
        top: Option<CollisionTile>,
        top_right: Option<CollisionTile>,
        right: Option<CollisionTile>,
        bottom_right: Option<CollisionTile>,
        bottom: Option<CollisionTile>,
        bottom_left: Option<CollisionTile>,
        left: Option<CollisionTile>,
    ) -> GraphicTile {
        GraphicTile::from_neighbours(
            top_left.unwrap_or_else(CollisionTile::full),
            top.unwrap_or_else(CollisionTile::full),
            top_right.unwrap_or_else(CollisionTile::full),
            right.unwrap_or_else(CollisionTile::full),
            bottom_right.unwrap_or_else(CollisionTile::full),
            bottom.unwrap_or_else(CollisionTile::full),
            bottom_left.unwrap_or_else(CollisionTile::full),
            left.unwrap_or_else(CollisionTile::full),
        )
    }

    /// Creates the correct graphic tile depending on the neighbours.
    pub fn from_neighbours(
        top_left: CollisionTile,
        top: CollisionTile,
        top_right: CollisionTile,
        right: CollisionTile,
        bottom_right: CollisionTile,
        bottom: CollisionTile,
        bottom_left: CollisionTile,
        left: CollisionTile,
    ) -> GraphicTile {

        let mut byte = 0;

        if top_left != CollisionTile::full() ||
           left     != CollisionTile::full() ||
           top      != CollisionTile::full() {
            byte += 1;
        }

        if top != CollisionTile::full() {
            byte += 2;
        }

        if top_right != CollisionTile::full() ||
           top       != CollisionTile::full() ||
           right     != CollisionTile::full() {
            byte += 4;
        }

        if right != CollisionTile::full() {
            byte += 8;
        }

        if bottom_right != CollisionTile::full() ||
           bottom       != CollisionTile::full() ||
           right        != CollisionTile::full() {
            byte += 16;
        }

        if bottom != CollisionTile::full() {
            byte += 32;
        }

        if bottom_left != CollisionTile::full() ||
           bottom      != CollisionTile::full() ||
           left        != CollisionTile::full() {
            byte += 64;
        }

        if left != CollisionTile::full() {
            byte += 128;
        }

        GraphicTile(Some(byte_to_index(byte)))
    }

    /// Returns the offset to the corresponding graphic tile in the texture.
    pub fn offset(self) -> (i32, i32) {
        match self.0 {
            None => (0, 0),
            Some(v) => (32 * v, 0)
        }
    }

    /// Returns true if the tile is visible.
    pub fn is_visible(&self) -> bool {
        self.0.is_some()
    }
}

/// A tile and its position.
pub struct PositionedTile {
    /// The graphic representation of the positioned tile.
    pub graphic: GraphicTile,

    /// The collision representation of the positioned tile.
    pub collision: CollisionTile,

    /// The position of the positioned tile.
    pub position: (f32, f32),
}

impl Drawable for PositionedTile {
    fn texture(&self) -> Texture {
        Texture::Overworld
    }

    fn texture_rect(&self) -> IntRect {
        let offset = self.graphic.offset();
        IntRect::new(offset.0, offset.1, SPRITE_SIZE_I32, SPRITE_SIZE_I32)
    }

    fn position(&self) -> Vector2<f32> {
        self.position.into()
    }
}

/// The map represents the tiles contained in a level.
#[derive(Serialize, Deserialize)]
pub struct Map {
    /// The entrace point of the character in the map.
    entrance: (usize, usize),

    /// The tiles contained in the level.
    tiles: Matrix<(CollisionTile, GraphicTile)>,
}

impl Map {
    /// Creates a map full of nothing, with a ground at the bottom.
    pub fn new(rows: usize, cols: usize) -> Map {
        let mut tiles = Matrix::from_size(rows, cols, CollisionTile::empty());
        let rows = tiles.rows();

        for i in 0..tiles.cols() {
            tiles[(rows - 1, i)] = CollisionTile::full();
        }

        Map::from_collision_tiles(tiles)
    }

    /// Encodes the map into a binary format.
    pub fn encode(&self) -> Result<Vec<u8>> {
        serialize(&self).map_err(Error::Encoding)
    }

    /// Decodes a map from bytes.
    pub fn decode(content: &[u8]) -> Result<Map> {
        deserialize(content).map_err(Error::Decoding)
    }

    /// Saves the map to a file.
    pub fn save<P: AsRef<Path>>(&self, path: P) -> Result<()> {
        let file = File::create(path.as_ref()).map_err(Error::Save)?;
        let mut writer = BufWriter::new(file);
        serialize_into(&mut writer, &self).map_err(Error::Encoding)?;
        Ok(())
    }

    /// Loads a map from a file.
    pub fn load<P: AsRef<Path>>(path: P) -> Result<Map> {
        let file = File::open(path.as_ref()).map_err(Error::Load)?;
        let mut reader = BufReader::new(file);
        Ok(deserialize_from(&mut reader).map_err(Error::Decoding)?)
    }

    /// Creates a map from its tiles.
    pub fn from_collision_tiles(tiles: Matrix<CollisionTile>) -> Map {
        let rows = tiles.rows();
        let cols = tiles.cols();

        let mut matrix =
            Matrix::from_size(rows, cols, (CollisionTile::empty(), GraphicTile(None)));

        for i in 0..rows {
            for j in 0..cols {
                let graphic = if tiles[(i, j)] == CollisionTile::full() {
                    // TODO This is uggly
                    // If there is an overflow, we should give None instead
                    let (i, j) = (i as isize, j as isize);

                    GraphicTile::from_neighbour_options(
                        tiles.get(((i - 1) as usize, (j - 1) as usize)).cloned(),
                        tiles.get(((i    ) as usize, (j - 1) as usize)).cloned(),
                        tiles.get(((i + 1) as usize, (j - 1) as usize)).cloned(),
                        tiles.get(((i + 1) as usize, (j    ) as usize)).cloned(),
                        tiles.get(((i + 1) as usize, (j + 1) as usize)).cloned(),
                        tiles.get(((i    ) as usize, (j + 1) as usize)).cloned(),
                        tiles.get(((i - 1) as usize, (j + 1) as usize)).cloned(),
                        tiles.get(((i - 1) as usize, (j    ) as usize)).cloned(),
                    )
                } else {
                    GraphicTile(None)
                };

                matrix[(i, j)] = (tiles[(i, j)], graphic);
            }
        }

        Map {
            entrance: Map::find_entrance(&matrix),
            tiles: matrix,
        }
    }

    /// Returns the graphic tile corresponding to the collision tiles.
    pub fn graphic_tile(&self, i: usize, j: usize) -> GraphicTile {
        if self.tiles[(i, j)].0 == CollisionTile::full() {
            GraphicTile::from_neighbour_options(
                self.tiles.get(((i - 1) as usize, (j - 1) as usize)).map(|x| x.0),
                self.tiles.get(((i - 1) as usize, (j    ) as usize)).map(|x| x.0),
                self.tiles.get(((i - 1) as usize, (j + 1) as usize)).map(|x| x.0),
                self.tiles.get(((i    ) as usize, (j + 1) as usize)).map(|x| x.0),
                self.tiles.get(((i + 1) as usize, (j + 1) as usize)).map(|x| x.0),
                self.tiles.get(((i + 1) as usize, (j    ) as usize)).map(|x| x.0),
                self.tiles.get(((i + 1) as usize, (j - 1) as usize)).map(|x| x.0),
                self.tiles.get(((i    ) as usize, (j - 1) as usize)).map(|x| x.0),
            )
        } else {
            GraphicTile(None)
        }
    }

    /// Returns a tile of the map.
    pub fn tile(&self, i: usize, j: usize) -> Option<CollisionTile> {
        self.tiles.get((i, j)).map(|x| x.0)
    }

    /// Changes a tile of the map.
    pub fn set_tile(&mut self, i: usize, j: usize, tile: CollisionTile) {
        self.tiles[(i, j)] = (tile, GraphicTile(None));
        for i in (i - 1) ..= (i + 1) {
            for j in (j - 1) ..= (j + 1) {
                if self.tiles.get((i, j)).is_some() {
                    self.tiles[(i, j)].1 = self.graphic_tile(i, j);
                }
            }
        }
    }

    /// Finds a possible entrance.
    pub fn find_entrance(tiles: &Matrix<(CollisionTile, GraphicTile)>) -> (usize, usize) {
        (tiles.rows() - 5, 1)
    }

    /// Returns the entrance of the map.
    pub fn entrance(&self) -> (usize, usize) {
        self.entrance
    }

    /// Returns an iterator to the positioned tiles.
    pub fn at(&self, row: usize, col: usize) -> PositionedTile {
        PositionedTile {
            collision: self.tiles[(row, col)].0,
            graphic: self.tiles[(row, col)].1,
            position: (col as f32 * SPRITE_SIZE_F32, row as f32 * SPRITE_SIZE_F32),
        }
    }

    /// Returns the number of rows of the map.
    pub fn rows(&self) -> usize {
        self.tiles.rows()
    }

    /// Returns the number of columns of the map.
    pub fn cols(&self) -> usize {
        self.tiles.cols()
    }

    /// Checks whether the bounding box collides with elements of the map.
    ///
    /// Returns the new correct position.
    pub fn collides_bbox(
        &self,
        old: FloatRect,
        new: FloatRect,
    ) -> Option<(CollisionAxis, Vector2<f32>, Damage)> {

        let mut damage = Damage::None;

        let cols = self.tiles.cols() - 1;
        let rows = self.tiles.rows() - 1;

        let min_col = clamp(new.left / SPRITE_SIZE_F32, 0.0, cols as f32) as usize;
        let min_row = clamp(new.top / SPRITE_SIZE_F32, 0.0, rows as f32) as usize;

        let max_col = clamp((new.left + new.width) / SPRITE_SIZE_F32, 0.0, cols as f32) as usize;
        let max_row = clamp((new.top + new.height) / SPRITE_SIZE_F32, 0.0, rows as f32) as usize;

        let mut collision_x = false;
        let mut collision_y = false;
        let mut new = new;

        for col in min_col..=max_col {
            for row in min_row..=max_row {
                let tile_left = col as f32 * SPRITE_SIZE_F32;
                let tile_top = row as f32 * SPRITE_SIZE_F32;

                let tile = FloatRect::new(tile_left, tile_top, SPRITE_SIZE_F32, SPRITE_SIZE_F32);

                if !overlap(new, tile) {
                    continue;
                }

                // Collisions between feet and ground
                if self.tiles[(row, col)].0.from_top
                    && old.top + old.height <= tile_top
                    && new.top + new.height >= tile_top
                {
                    collision_y = true;
                    new.top = tile_top - new.height;
                }

                if !overlap(new, tile) {
                    continue;
                }

                // Collisions between right and right wall
                if self.tiles[(row, col)].0.from_left
                    && old.left + old.width <= tile_left
                    && new.left + new.width >= tile_left
                {
                    collision_x = true;
                    new.left = tile_left - new.width;
                }

                if !overlap(new, tile) {
                    continue;
                }

                // Collisions between left and left wall
                if self.tiles[(row, col)].0.from_right
                    && old.left >= tile_left + SPRITE_SIZE_F32
                    && new.left <= tile_left + SPRITE_SIZE_F32
                {
                    collision_x = true;
                    new.left = tile_left + SPRITE_SIZE_F32;
                }

                if !overlap(new, tile) {
                    continue;
                }

                // Collisions between head and roof
                if self.tiles[(row, col)].0.from_bottom
                    && old.top >= tile_top + SPRITE_SIZE_F32
                    && new.top <= tile_top + SPRITE_SIZE_F32
                {
                    collision_y = true;
                    new.top = tile_top + SPRITE_SIZE_F32;
                }
            }
        }

        // Collision between the player and left border of the level
        if new.left < 0.0 {
            new.left = 0.0;
            collision_x = true;
        }

        // Collision between the player and right border of the level
        if new.left > cols as f32 * SPRITE_SIZE_F32 {
            new.left = cols as f32 * SPRITE_SIZE_F32;
            collision_x = true;
        }

        // Collision between the player and the void
        if new.top > self.tiles.rows() as f32 * SPRITE_SIZE_F32 {
            new.top = self.tiles.rows() as f32 * SPRITE_SIZE_F32;
            collision_y = true;
            damage = Damage::Death;
        }

        let new_pos = Vector2::new(new.left, new.top);
        match (collision_x, collision_y) {
            (true, true) => Some((CollisionAxis::Both, new_pos, damage)),
            (true, false) => Some((CollisionAxis::X, new_pos, damage)),
            (false, true) => Some((CollisionAxis::Y, new_pos, damage)),
            (false, false) => None,
        }
    }
}

/// Checks if two boxes overlap.
pub fn overlap(box1: FloatRect, box2: FloatRect) -> bool {
    box2.left < box1.left + box1.width
        && box2.left + box2.width > box1.left
        && box2.top < box1.top + box1.height
        && box2.top + box2.height > box1.top
}
